import json
import logging
import os
import os.path
import pathlib
import stat
import tarfile
import zipfile
from typing import List

import toml

logger = logging.getLogger("utils")


def untar(archive_path, destination_folder):
    logger.debug(f"untar [{archive_path}] to [{destination_folder}].")

    ensure_folder(destination_folder)
    tar = tarfile.open(archive_path)
    tar.extractall(path=destination_folder)
    tar.close()

    logger.debug("untar done.")


def unzip(archive_path, destination_folder):
    logger.debug(f"unzip [{archive_path}] to [{destination_folder}].")

    ensure_folder(destination_folder)
    with zipfile.ZipFile(archive_path, "r") as my_zip:
        my_zip.extractall(destination_folder)

    logger.debug("unzip done.")


def ensure_folder(folder):
    pathlib.Path(folder).mkdir(parents=True, exist_ok=True)


def read_lines(file):
    with open(file) as f:
        lines = f.readlines()
    lines = [line.strip() for line in lines]
    lines = [line for line in lines if line]
    return lines


def read_file(filename):
    with open(filename) as f:
        return f.read()


def write_file(filename, text):
    with open(filename, "w") as f:
        return f.write(text)


def read_toml_file(filename):
    return toml.load(filename)


def write_toml_file(filename, data):
    with open(filename, "w") as f:
        toml.dump(data, f)


def read_json_file(filename):
    with open(filename) as f:
        return json.load(f)


def dump_out_json(data, outfile):
    json.dump(data, outfile, indent=4)


def get_subfolders(folder):
    return [item.name for item in os.scandir(folder) if item.is_dir() and not item.name.startswith(".")]


def mark_executable(file):
    logger.debug(f"Mark [{file}] as executable")
    st = os.stat(file)
    os.chmod(file, st.st_mode | stat.S_IEXEC)


def find_in_dictionary(dictionary, compound_path):
    keys = compound_path.split(".")
    node = dictionary
    for key in keys:
        node = node.get(key)
        if node is None:
            break

    return node


def list_files(folder: str) -> List[str]:
    files = os.listdir(folder)
    files = [os.path.join(folder, f) for f in files]
    return files



import logging

from erdpy.proxy.http_facade import do_get, do_post

METACHAIN_ID = 4294967295
ANY_SHARD_ID = 0

logger = logging.getLogger("proxy")


class ElrondProxy:
    def __init__(self, url):
        self.url = url

    def get_account_nonce(self, address):
        url = f"{self.url}/address/{address.bech32()}"
        response = do_get(url)
        nonce = response["account"]["nonce"]
        return nonce

    def get_account_balance(self, address):
        url = f"{self.url}/address/{address.bech32()}/balance"
        response = do_get(url)
        balance = response["balance"]
        return balance

    def get_account(self, address):
        url = f"{self.url}/address/{address.bech32()}"
        response = do_get(url)
        return response

    def get_account_transactions(self, address):
        url = f"{self.url}/address/{address.bech32()}/transactions"
        response = do_get(url)
        return response

    def get_num_shards(self):
        url = f"{self.url}/network/config"
        response = do_get(url)
        num_shards_without_meta = response["config"]["erd_num_shards_without_meta"] or 0
        return num_shards_without_meta + 1

    def get_last_block_nonce(self, shard_id):
        if shard_id == "metachain":
            metrics = self._get_network_status(METACHAIN_ID)
        else:
            metrics = self._get_network_status(shard_id)

        nonce = metrics["erd_nonce"]
        return nonce

    def get_meta_nonce(self):
        url = f"{self.url}/block/meta-nonce"
        response = do_get(url)
        nonce = response["nonce"]
        return nonce

    def get_meta_block(self, nonce):
        url = f"{self.url}/block/meta/{nonce}"
        response = do_get(url)
        nonce = response["block"]
        return nonce

    def get_gas_price(self):
        metrics = self._get_network_config()
        price = metrics["erd_min_gas_price"]
        return price

    def get_chain_id(self):
        metrics = self._get_network_config()
        chain_id = metrics["erd_chain_id"]
        return chain_id

    def _get_network_status(self, shard_id):
        url = f"{self.url}/network/status/{shard_id}"
        response = do_get(url)
        details = response["status"]
        return details

    def _get_network_config(self):
        url = f"{self.url}/network/config"
        response = do_get(url)
        details = response["config"]
        return details

    def send_transaction(self, payload):
        url = f"{self.url}/transaction/send"
        response = do_post(url, payload)
        tx_hash = response["txHash"]
        return tx_hash

    def send_transactions(self, payload):
        url = f"{self.url}/transaction/send-multiple"
        response = do_post(url, payload)
        # Proxy and Observers have different response format:
        num_sent = response.get("numOfSentTxs", 0) or response.get("txsSent", 0)
        hashes = response.get("txsHashes", None)
        return num_sent, hashes

    def query_contract(self, payload):
        url = f"{self.url}/vm-values/query"
        response = do_post(url, payload)
        return response




class KnownError(Exception):
    inner = None

    def __init__(self, message, inner=None):
        super().__init__(message)
        self.inner = inner

    def get_pretty(self) -> str:
        if self.inner:
            return f"""{self}
... {self.inner}
"""
        return str(self)


class ProgrammingError(KnownError):
    pass


class TemplateMissingError(KnownError):
    def __init__(self, template):
        super().__init__(f"Template missing: {template}")


class DownloadError(KnownError):
    pass


class BadUrlError(DownloadError):
    def __init__(self):
        pass


class DependencyMissing(KnownError):
    def __init__(self, group):
        super().__init__(f"Dependency missing. Group: {group}")


class BadDirectory(KnownError):
    def __init__(self, directory):
        super().__init__(f"Bad directory: {directory}")


class NotSupportedProject(KnownError):
    def __init__(self, directory):
        super().__init__(f"Directory is not a supported project: {directory}")


class PlatformNotSupported(KnownError):
    def __init__(self, action_or_item, platform):
        super().__init__(f"[{action_or_item}] is not supported on platform [{platform}].")


class BuildError(KnownError):
    def __init__(self, message, inner=None):
        super().__init__(f"Build error: {message}.", inner)


class BadSink(ProgrammingError):
    def __init__(self, name):
        super().__init__(f"Bad sink:\n {name}.")


class UnknownArgumentFormat(KnownError):
    def __init__(self, argument):
        super().__init__(f"Cannot handle non-hex, non-number arguments yet: {argument}.")


class ProxyRequestError(KnownError):
    def __init__(self, url, data):
        super().__init__(f"Proxy request error for url [{url}]: {data}")


class BlockAtlasRequestError(KnownError):
    def __init__(self, url, data):
        super().__init__(f"Block Atlas request error for url [{url}]: {data}")


class BadInputError(KnownError):
    def __init__(self, input, message):
        super().__init__(f"Bad input [{input}]: {message}")


class BadAddressFormatError(KnownError):
    def __init__(self, value):
        super().__init__(f"Bad address [{value}].")


class EmptyAddressError(KnownError):
    def __init__(self):
        super().__init__(f"Address is empty.")


class ExternalProcessError(KnownError):
    def __init__(self, command_line, message):
        super().__init__(f"""External process error:
Command line: {command_line}
Output: {message}""")



import os
import sys
from argparse import FileType

from erdpy import config, facade, ide, projects


def setup_parser(subparsers):
    parser = subparsers.add_parser("contract")
    subparsers = parser.add_subparsers()

    sub = subparsers.add_parser("new", description="Create a new Smart Contract project based on a template.")
    sub.add_argument("name")
    sub.add_argument("--template", required=True, help="the template to use")
    sub.add_argument("--directory", type=str, default=os.getcwd(), help="the parent directory of the project")
    sub.set_defaults(func=create)

    sub = subparsers.add_parser("templates", description="List the available Smart Contract templates.")
    sub.add_argument("--json", action="store_true", help="whether to print the list in JSON format")
    sub.set_defaults(func=list_templates)

    sub = subparsers.add_parser("build", description="Build a Smart Contract project using the appropriate buildchain.")
    sub.add_argument("project", nargs='?', default=os.getcwd())
    sub.add_argument("--debug", action="store_true", default=False)
    sub.add_argument("--no-optimization", action="store_true", default=False)
    sub.set_defaults(func=build)

    sub = subparsers.add_parser("deploy", description="Deploy a Smart Contract.")
    sub.add_argument("project", nargs='?', default=os.getcwd(), help="the project directory")
    sub.add_argument("--proxy", required=True, help="the URL of the proxy")
    sub.add_argument("--pem", required=True, help="the PEM file of the owner")
    sub.add_argument("--arguments", nargs='+', help="constructor arguments")
    sub.add_argument("--gas-price", default=config.DEFAULT_GAS_PRICE, help="the gas price")
    sub.add_argument("--gas-limit", required=True, help="the gas limit")
    sub.add_argument("--value", default="0", help="the value to transfer")
    sub.add_argument("--metadata-upgradeable", action="store_true", default=False, help="whether the contract is upgradeable")
    sub.add_argument("--outfile", type=FileType("w"), default=sys.stdout, help="where to save the command's output")
    sub.set_defaults(func=deploy)

    sub = subparsers.add_parser("call")
    sub.add_argument("contract")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--pem", required=True)
    sub.add_argument("--function", required=True)
    sub.add_argument("--arguments", nargs='+')
    sub.add_argument("--gas-price", default=config.DEFAULT_GAS_PRICE)
    sub.add_argument("--gas-limit", required=True)
    sub.add_argument("--value", default="0")
    sub.set_defaults(func=call)

    sub = subparsers.add_parser("upgrade")
    sub.add_argument("contract")
    sub.add_argument("project")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--pem", required=True)
    sub.add_argument("--arguments", nargs='+')
    sub.add_argument("--gas-price", default=config.DEFAULT_GAS_PRICE)
    sub.add_argument("--gas-limit", required=True)
    sub.add_argument("--value", default="0")
    sub.add_argument("--metadata-upgradeable", action="store_true", default=False)
    sub.set_defaults(func=upgrade)

    sub = subparsers.add_parser("query")
    sub.add_argument("contract")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--function", required=True)
    sub.add_argument("--arguments", nargs='+')
    sub.set_defaults(func=query)

    # TODO: arwendebug
    # node_parser = subparsers.add_parser("nodedebug")
    # group = node_parser.add_mutually_exclusive_group()
    # group.add_argument('--stop', action='store_true')
    # group.add_argument('--restart', action='store_true', default=True)
    # node_parser.set_defaults(func=do_nodedebug)

    sub = subparsers.add_parser("test")
    sub.add_argument("project", nargs='?', default=os.getcwd())
    sub.add_argument("--directory", default="test")
    sub.add_argument("--wildcard", required=False)
    sub.set_defaults(func=run_tests)

    sub = subparsers.add_parser("ide")
    sub.add_argument("workspace", nargs='?', default=os.getcwd())
    sub.set_defaults(func=run_ide)


def list_templates(args):
    json = args.json
    projects.list_project_templates(json)


def create(args):
    name = args.name
    template = args.template
    directory = args.directory

    projects.create_from_template(name, template, directory)


def build(args):
    project = args.project
    options = {
        "debug": args.debug,
        "optimized": not args.no_optimization,
        "verbose": args.verbose
    }

    projects.build_project(project, options)


def deploy(args):
    facade.deploy_smart_contract(args)


def call(args):
    facade.call_smart_contract(args)


def upgrade(args):
    facade.upgrade_smart_contract(args)


def query(args):
    facade.query_smart_contract(args)


def run_tests(args):
    projects.run_tests(args)


def run_ide(args):
    workspace = args.workspace
    ide.run_ide(workspace)


# def do_nodedebug(args):
#     stop = args.stop
#     restart = args.restart

#     if restart:
#         nodedebug.stop()
#         nodedebug.start()
#     elif stop:
#         nodedebug.stop()
#     else:
#         nodedebug.start()



import asyncio
import logging
import traceback

from erdpy import errors, nodedebug
from erdpy.proxy import ElrondProxy

logger = logging.getLogger("environments")


class Environment:
    def __init__(self):
        pass

    def run_flow(self):
        raise NotImplementedError()

    def deploy_contract(self, contract, owner, arguments=None, gas_price=None, gas_limit=None):
        raise NotImplementedError()

    def execute_contract(self, contract, caller, function, arguments=None, gas_price=None, gas_limit=None):
        raise NotImplementedError()

    def query_contract(self, contract, function, arguments=None):
        raise NotImplementedError()


class DebugEnvironment(Environment):
    def __init__(self):
        super().__init__()

    def run_flow(self, flow):
        nodedebug.install_if_missing()

        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._run_node_debug_and_flow(flow))
        loop.close()
        asyncio.set_event_loop(asyncio.new_event_loop())

    async def _run_node_debug_and_flow(self, flow):
        await asyncio.wait([
            nodedebug.start_async(),
            self._wrap_flow(flow)
        ])

    async def _wrap_flow(self, flow):
        try:
            logger.debug("Wait before starting flow...")
            await asyncio.sleep(1)
            logger.debug("Starting flow...")
            flow()
            logger.debug("Flow ran.")
            await asyncio.sleep(1)
        except errors.KnownError as err:
            logger.critical(err)
        except Exception:
            print(traceback.format_exc())
        finally:
            nodedebug.stop()

    def deploy_contract(self, contract, owner, arguments=None, gas_price=None, gas_limit=None):
        raise Exception("Temporarily disabled: deploy_contract on debug")

        logger.debug("deploy_contract")
        tx_hash, contract_address = nodedebug.deploy(contract.bytecode, owner, arguments, gas_price, gas_limit)
        contract.address = contract_address
        return tx_hash, contract_address

    def execute_contract(self, contract, caller, function, arguments=None, gas_price=None, gas_limit=None):
        raise Exception("Temporarily disabled: execute_contract on debug")

        logger.debug("execute_contract")
        nodedebug.execute(contract.address, caller, function, arguments, gas_price, gas_limit)

    def query_contract(self, contract, function, arguments=None):
        raise Exception("Temporarily disabled: query_contract on debug")

        logger.debug("query_contract")
        return nodedebug.query(contract.address, function, arguments)


class TestnetEnvironment(Environment):
    def __init__(self, url):
        super().__init__()
        self.url = url

    def run_flow(self, flow):
        return self._wrap_flow(flow)

    def _wrap_flow(self, flow):
        try:
            logger.debug("Starting flow...")
            result = flow()
            logger.debug("Flow ran.")
            return result
        except errors.KnownError as err:
            logger.critical(err)
        except Exception:
            print(traceback.format_exc())

    def deploy_contract(self, contract, owner, arguments=None, gas_price=None, gas_limit=None, value=None):
        logger.debug("deploy_contract")
        proxy = self._get_proxy()
        tx_hash, address = contract.deploy(proxy, owner, arguments, gas_price, gas_limit, value)
        return tx_hash, contract.address

    def execute_contract(self, contract, caller, function, arguments=None, gas_price=None, gas_limit=None, value=None):
        logger.debug("execute_contract: %s", contract.address.bech32())
        proxy = self._get_proxy()
        tx_hash = contract.execute(proxy, caller, function, arguments, gas_price, gas_limit, value)
        return tx_hash

    def upgrade_contract(self, contract, caller, arguments=None, gas_price=None, gas_limit=None, value=None):
        logger.debug("upgrade_contract: %s", contract.address.bech32())
        proxy = self._get_proxy()
        tx_hash = contract.upgrade(proxy, caller, arguments, gas_price, gas_limit, value)
        return tx_hash

    def query_contract(self, contract, function, arguments=None):
        logger.debug("query_contract: %s", contract.address.bech32())
        proxy = self._get_proxy()
        return_data = contract.query(proxy, function, arguments)
        return return_data

    def _get_proxy(self):
        return ElrondProxy(self.url)



import logging
import sys
from argparse import ArgumentParser

from erdpy import (cli_contracts, cli_install, cli_transactions,
                   cli_validators, config, errors, facade, proxy)
from erdpy._version import __version__

logger = logging.getLogger("cli")


def main():
    try:
        _do_main()
    except errors.KnownError as err:
        logger.fatal(err.get_pretty())
        sys.exit(1)


def _do_main():
    parser = setup_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)

    if not hasattr(args, "func"):
        parser.print_help()
    else:
        args.func(args)


def setup_parser():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()

    parser.add_argument('-v', '--version', action='version', version=f"erdpy {__version__}")
    parser.add_argument("--verbose", action="store_true", default=False)

    cli_install.setup_parser(subparsers)
    cli_contracts.setup_parser(subparsers)
    cli_validators.setup_parser(subparsers)
    cli_transactions.setup_parser(subparsers)

    setup_parser_accounts(subparsers)
    setup_parser_wallet(subparsers)
    setup_parser_cost(subparsers)
    setup_parser_network(subparsers)
    setup_parser_blockatlas(subparsers)
    setup_parser_dispatcher(subparsers)

    return parser


def setup_parser_accounts(subparsers):
    parser = subparsers.add_parser("account")
    subparsers = parser.add_subparsers()

    sub = subparsers.add_parser("get")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--address", required=True)
    sub.add_argument("--balance", required=False, nargs='?', const=True, default=False)
    sub.add_argument("--nonce", required=False, nargs='?', const=True, default=False)
    sub.set_defaults(func=get_account)

    sub = subparsers.add_parser("get-transactions")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--address", required=True)
    sub.set_defaults(func=get_account_transactions)


def setup_parser_wallet(subparsers):
    wallet_parser = subparsers.add_parser("wallet")
    wallet_subparsers = wallet_parser.add_subparsers()

    subparser = wallet_subparsers.add_parser("generate")
    subparser.add_argument("pem")
    subparser.add_argument("--mnemonic", action="store_true")
    subparser.set_defaults(func=generate_pem)

    subparser = wallet_subparsers.add_parser("bech32")
    subparser.add_argument("value")
    group = subparser.add_mutually_exclusive_group(required=True)
    group.add_argument("--encode", action="store_true")
    group.add_argument("--decode", action="store_true")
    subparser.set_defaults(func=do_bech32)


def setup_parser_cost(subparsers):
    cost_parser = subparsers.add_parser("cost")
    cost_subparsers = cost_parser.add_subparsers()

    sub = cost_subparsers.add_parser("gas-price")
    sub.add_argument("--proxy", required=True)
    sub.set_defaults(func=get_gas_price)

    sub = cost_subparsers.add_parser("transaction")
    tx_types = [proxy.TxTypes.SC_CALL, proxy.TxTypes.MOVE_BALANCE, proxy.TxTypes.SC_DEPLOY]
    sub.add_argument("tx_type", choices=tx_types)
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--data", required=False)
    sub.add_argument("--sc-address", required=False)
    sub.add_argument("--sc-path", required=False)
    sub.add_argument("--function", required=False)
    sub.add_argument("--arguments", nargs='+', required=False)
    sub.set_defaults(func=get_transaction_cost)


def setup_parser_network(subparsers):
    parser = subparsers.add_parser("network")
    subparsers = parser.add_subparsers()

    sub = subparsers.add_parser("num-shards")
    sub.add_argument("--proxy", required=True)
    sub.set_defaults(func=get_num_shards)

    sub = subparsers.add_parser("last-block-nonce")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--shard-id", required=True)
    sub.set_defaults(func=get_last_block_nonce)

    sub = subparsers.add_parser("chain-id")
    sub.add_argument("--proxy", required=True)
    sub.set_defaults(func=get_chain_id)

    sub = subparsers.add_parser("meta-nonce")
    sub.add_argument("--proxy", required=True)
    sub.set_defaults(func=get_meta_nonce)

    sub = subparsers.add_parser("meta-block")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--nonce", required=True, type=int)
    sub.set_defaults(func=get_meta_block)


def setup_parser_blockatlas(subparsers):
    parser = subparsers.add_parser("blockatlas")
    subparsers = parser.add_subparsers()

    parser.add_argument("--url", required=True)
    parser.add_argument("--coin", required=True)

    sub = subparsers.add_parser("current-block-number")
    sub.set_defaults(func=facade.blockatlas_get_current_block_number)

    sub = subparsers.add_parser("block-by-number")
    sub.add_argument("--number", required=True)
    sub.set_defaults(func=facade.blockatlas_get_block_by_number)

    sub = subparsers.add_parser("transactions")
    sub.add_argument("--address", required=True)
    sub.set_defaults(func=facade.blockatlas_get_txs_by_address)


def setup_parser_dispatcher(subparsers):
    parser = subparsers.add_parser("dispatcher")
    subparsers = parser.add_subparsers()

    sub = subparsers.add_parser("enqueue")
    sub.add_argument("--value", default="0")
    sub.add_argument("--receiver", required=True)
    sub.add_argument("--gas-price", default=config.DEFAULT_GAS_PRICE)
    sub.add_argument("--gas-limit", required=True)
    sub.add_argument("--data", default="")
    sub.set_defaults(func=enqueue_transaction)

    sub = subparsers.add_parser("dispatch")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--pem", required=True)
    sub.set_defaults(func=dispatch_transactions)

    sub = subparsers.add_parser("dispatch-continuously")
    sub.add_argument("--proxy", required=True)
    sub.add_argument("--pem", required=True)
    sub.add_argument("--interval", required=True)
    sub.set_defaults(func=dispatch_transactions_continuously)

    sub = subparsers.add_parser("clean")
    sub.set_defaults(func=clean_transactions_queue)


def get_account(args):
    if args.balance:
        facade.get_account_balance(args)
    elif args.nonce:
        facade.get_account_nonce(args)
    else:
        facade.get_account(args)


def get_account_transactions(args):
    facade.get_account_transactions(args)


def get_transaction_cost(args):
    facade.get_transaction_cost(args)


def get_num_shards(args):
    facade.get_num_shards(args)


def get_last_block_nonce(args):
    facade.get_last_block_nonce(args)


def get_gas_price(args):
    facade.get_gas_price(args)


def get_chain_id(args):
    facade.get_chain_id(args)


def get_meta_nonce(args):
    facade.get_meta_nonce(args)


def get_meta_block(args):
    facade.get_meta_block(args)


def generate_pem(args):
    facade.generate_pem(args)


def do_bech32(args):
    facade.do_bech32(args)


def enqueue_transaction(args):
    facade.enqueue_transaction(args)


def dispatch_transactions(args):
    facade.dispatch_transactions(args)


def dispatch_transactions_continuously(args):
    facade.dispatch_transactions_continuously(args)


def clean_transactions_queue(args):
    facade.clean_transactions_queue()


if __name__ == "__main__":
    main()




import logging
import os
from os import path

from erdpy import config, downloader, workstation, errors, myprocess, utils

logger = logging.getLogger("modules")


class DependencyModule:
    def __init__(self, key, name, tag, groups):
        self.key = key
        self.name = name
        self.tag = tag
        self.groups = groups

    def install(self, overwrite):
        raise NotImplementedError()

    def uninstall(self):
        raise NotImplementedError()

    def is_installed(self):
        raise NotImplementedError()

    def get_env(self):
        raise NotImplementedError()


class StandaloneModule(DependencyModule):
    def __init__(self, key, name, tag, groups, urls_by_platform):
        super().__init__(key, name, tag, groups)
        self.urls_by_platform = urls_by_platform

    def install(self, overwrite):
        logger.debug(f"install: name={self.name}, tag={self.tag}")

        if self._should_skip(overwrite):
            logger.debug("Already exists. Skip install.")
            return

        self._download()
        self._extract()

    def _should_skip(self, overwrite):
        if overwrite:
            return False

        return self.is_installed()

    def is_installed(self):
        return path.isdir(self.get_directory())

    def _download(self):
        url = self._get_download_url()
        archive_path = self._get_archive_path()
        downloader.download(url, archive_path)

    def _extract(self):
        archive_path = self._get_archive_path()
        destination_folder = self.get_directory()
        utils.untar(archive_path, destination_folder)

    def get_directory(self):
        tools_folder = workstation.get_tools_folder()
        folder = path.join(tools_folder, self.name, self.tag)
        return folder

    def _get_download_url(self):
        platform = workstation.get_platform()
        url = self.urls_by_platform.get(platform)

        if url is None:
            raise errors.PlatformNotSupported(self.name, platform)

        url = f"{config.DOWNLOAD_MIRROR}/{url}"
        return url

    def _get_archive_path(self):
        tools_folder = workstation.get_tools_folder()
        archive = path.join(tools_folder, f"{self.name}.{self.tag}.tar.gz")
        return archive

    def get_env(self):
        return {
            "LD_LIBRARY_PATH": f"{path.join(self.get_directory())}:{os.environ.get('LD_LIBRARY_PATH')}"
        }

# TODO: clang, cpp modules seem to require "sudo apt install libtinfo5"
# clang-9: error while loading shared libraries: libtinfo.so.5: cannot open shared object file: No such file or directory


class SOLLModule(StandaloneModule):
    def __init__(self, key, name, tag, groups, urls_by_platform):
        super().__init__(key, name, tag, groups, urls_by_platform)

    def install(self, overwrite):
        super().install(overwrite)
        utils.mark_executable(path.join(self.get_directory(), "soll"))


class ArwenToolsModule(StandaloneModule):
    def __init__(self, key, name, tag, groups, urls_by_platform):
        super().__init__(key, name, tag, groups, urls_by_platform)

    def install(self, overwrite):
        super().install(overwrite)
        utils.mark_executable(path.join(self.get_directory(), "arwen"))
        utils.mark_executable(path.join(self.get_directory(), "arwendebug"))
        utils.mark_executable(path.join(self.get_directory(), "test"))


class Rust(DependencyModule):
    def __init__(self, key, name, tag, groups):
        super().__init__(key, name, tag, groups)

    def install(self, overwrite):
        logger.debug(f"install: name={self.name}, tag={self.tag}")

        if self._should_skip(overwrite):
            logger.debug("Already exists. Skip install.")
            return

        rustup_path = self._get_rustup_path()
        downloader.download("https://sh.rustup.rs", rustup_path)
        utils.mark_executable(rustup_path)

        args = [rustup_path, "--verbose", "--default-toolchain", "nightly", "--profile",
                "minimal", "--target", "wasm32-unknown-unknown", "--no-modify-path", "-y"]
        myprocess.run_process_async(args, env=self.get_env())

    def _should_skip(self, overwrite):
        if overwrite:
            return False
        return self.is_installed()

    def is_installed(self):
        try:
            myprocess.run_process(["rustc", "--version"], env=self.get_env())
            return True
        except Exception:
            return False

    def _get_rustup_path(self):
        tools_folder = workstation.get_tools_folder()
        return path.join(tools_folder, "rustup.sh")

    def get_directory(self):
        tools_folder = workstation.get_tools_folder()
        return path.join(tools_folder, "vendor-rust")

    def get_env(self):
        return {
            "PATH": f"{path.join(self.get_directory(), 'bin')}:{os.environ['PATH']}",
            "RUSTUP_HOME": self.get_directory(),
            "CARGO_HOME": self.get_directory()
        }



import os
import sys

from erdpy import config, facade, transactions


def setup_parser(subparsers):
    sub = subparsers.add_parser("tx-prepare")
    _add_common_arguments(sub)
    sub.add_argument("--tag", default="untitled")
    sub.add_argument("workspace", nargs='?', default=os.getcwd())
    sub.set_defaults(func=tx_prepare)

    sub = subparsers.add_parser("tx-send")
    sub.add_argument("tx")
    sub.add_argument("--proxy", required=True)
    sub.set_defaults(func=tx_send)

    sub = subparsers.add_parser("tx-prepare-and-send")
    _add_common_arguments(sub)
    sub.add_argument("--proxy", required=True)
    sub.set_defaults(func=tx_prepare_and_send)


def _add_common_arguments(sub):
    sub.add_argument("--pem", required=True)
    sub.add_argument("--nonce", type=int, required=not("--recall-nonce" in sys.argv))
    sub.add_argument("--recall-nonce", action="store_true", default=False)
    sub.add_argument("--value", default="0")
    sub.add_argument("--receiver", required=True)
    sub.add_argument("--gas-price", default=config.DEFAULT_GAS_PRICE)
    sub.add_argument("--gas-limit", required=True)
    sub.add_argument("--data", default="")


def tx_prepare(args):
    transactions.prepare(args)


def tx_send(args):
    facade.send_prepared_transaction(args)


def tx_prepare_and_send(args):
    facade.prepare_and_send_transaction(args)



import json
import logging
import os
from os import path

from erdpy import dependencies, errors, utils
from erdpy.projects import shared
from erdpy.projects.project_rust import CargoFile
from erdpy.projects.templates_config import get_templates_repositories
from texttable import Texttable

logger = logging.getLogger("projects.templates")


def list_project_templates(as_json=False):
    templates = []

    for repository in get_templates_repositories():
        repository.download()
        for template in repository.get_templates():
            templates.append(TemplateSummary(template, repository))

    templates = sorted(templates, key=lambda item: item.name)

    if as_json:
        pretty_json = json.dumps([item.__dict__ for item in templates], indent=4)
        print(pretty_json)
    else:
        table = Texttable()
        table_data = [["Name", "Github", "Language"]]
        table_data.extend([[item.name, item.github, item.language] for item in templates])
        table.add_rows(table_data)
        print(table.draw())


class TemplateSummary():
    def __init__(self, name, repository):
        self.name = name
        self.github = repository.github
        self.language = repository.get_language(name)


def create_from_template(name, template_name, directory):
    directory = path.expanduser(directory)

    logger.info("create_from_template.name: %s", name)
    logger.info("create_from_template.template_name: %s", template_name)
    logger.info("create_from_template.directory: %s", directory)

    if not directory:
        logger.info("Using current directory")
        directory = os.getcwd()

    project_directory = path.join(directory, name)
    if path.exists(project_directory):
        raise errors.BadDirectory(project_directory)

    _download_templates_repositories()
    _copy_template(template_name, project_directory)

    template = _load_as_template(project_directory)
    template.apply(template_name, name)

    logger.info("Project created, template applied.")


def _download_templates_repositories():
    for repo in get_templates_repositories():
        repo.download()


def _copy_template(template, destination_path):
    for repo in get_templates_repositories():
        if repo.has_template(template):
            repo.copy_template(template, destination_path)
            return

    raise errors.TemplateMissingError(template)


def _load_as_template(directory):
    if shared.is_source_clang(directory):
        return TemplateClang(directory)
    if shared.is_source_sol(directory):
        return TemplateSol(directory)
    if shared.is_source_rust(directory):
        return TemplateRust(directory)


class Template:
    def __init__(self, directory):
        self.directory = directory

    def apply(self, template_name, project_name):
        self.template_name = template_name
        self.project_name = project_name
        self._extend()
        self._replace_placeholders()

    def _extend(self):
        pass

    def _replace_placeholders(self):
        pass


class TemplateClang(Template):
    pass


class TemplateRust(Template):
    def _extend(self):
        logger.info("TemplateRust._extend")

    def _replace_placeholders(self):
        rust_module = dependencies.get_module_by_key("rust")
        self.rust_directory = rust_module.get_directory()
        self.rust_bin_directory = path.join(self.rust_directory, "bin")

        cargo_path = path.join(self.directory, "Cargo.toml")
        cargo_debug_path = path.join(self.directory, "debug", "Cargo.toml")
        debug_main_path = path.join(self.directory, "debug", "src", "main.rs")
        test_dir_path = path.join(self.directory, "test")
        test_paths = utils.list_files(test_dir_path) if os.path.isdir(test_dir_path) else []

        logger.info("Updating cargo files...")

        cargo_file = CargoFile(cargo_path)
        cargo_file.package_name = self.project_name
        cargo_file.bin_name = self.project_name
        cargo_file.version = "0.0.1"
        cargo_file.authors = ["you"]
        cargo_file.edition = "2018"
        cargo_file.save()

        cargo_file_debug = CargoFile(cargo_debug_path)
        cargo_file_debug.package_name = f"{self.project_name}-debug"
        cargo_file_debug.version = "0.0.1"
        cargo_file_debug.authors = ["you"]
        cargo_file_debug.edition = "2018"
        cargo_file_debug.save()

        logger.info("Applying replacements...")

        self._replace_in_files(
            [debug_main_path],
            [
                # Example "use simple_coin::*" to "use my_project::*"
                (f"use {self.template_name.replace('-', '_')}::*", f"use {self.project_name.replace('-', '_')}::*")
            ]
        )

        self._replace_in_files(
            [cargo_debug_path],
            [
                (f"[dependencies.{self.template_name}]", f"[dependencies.{self.project_name}]")
            ]
        )

        self._replace_in_files(
            test_paths,
            [
                (f"{self.template_name}.wasm", f"{self.project_name}.wasm")
            ]
        )

    def _replace_in_files(self, files, replacements):
        for file in files:
            content = utils.read_file(file)

            for to_replace, replacement in replacements:
                content = content.replace(to_replace, replacement)

            utils.write_file(file, content)


class TemplateSol(Template):
    pass



# Configuration that should not be edited by the end user:
DOWNLOAD_MIRROR = "https://ide.elrond.com"
MODULES_CONFIG_URL = "https://raw.githubusercontent.com/ElrondNetwork/elrond-sdk/master/deps.json"
WITH_CHAIN_ID_AND_TX_VERSION = False
CHAIN_ID = ""
TX_VERSION = 0

# Configuration that may be edited by the end user:
ROOT_FOLDER_NAME = "elrondsdk"

DEFAULT_GAS_PRICE = 200000000000
GAS_PER_DATA_BYTE = 1500
MIN_GAS_LIMIT = 50000


class MetaChainSystemSCsCost:
    STAKE = 5000000
    UNSTAKE = 5000000
    UNBOND = 5000000
    CLAIM = 5000000
    GET = 5000000
    CHANGE_REWARD_ADDRESS = 5000000
    CHANGE_VALIDATOR_KEYS = 5000000
    UNJAIL = 5000000



import json
import logging
from collections import OrderedDict
from os import path

from erdpy import config, utils
from erdpy.accounts import Address
from erdpy.wallet import pem, signing

logger = logging.getLogger("transactions")


class PlainTransaction:
    def __init__(self):
        self.nonce = 0
        self.value = "0"
        self.sender = ""
        self.receiver = ""
        self.gasPrice = 0
        self.gasLimit = 0
        self.data = ""
        self.chainID = config.CHAIN_ID
        self.version = config.TX_VERSION

    def payload(self):
        return self.__dict__.copy()


class TransactionPayloadToSign:
    def __init__(self, transaction):
        self.__dict__.update(transaction.payload())

        # When signing the transaction, we base64 encode the "Data" field
        if transaction.data:
            self.data = transaction.data

    def to_json(self):
        ordered_fields = OrderedDict()
        ordered_fields["nonce"] = self.nonce
        ordered_fields["value"] = self.value
        ordered_fields["receiver"] = self.receiver
        ordered_fields["sender"] = self.sender
        ordered_fields["gasPrice"] = self.gasPrice
        ordered_fields["gasLimit"] = self.gasLimit

        if self.data:
            ordered_fields["data"] = self.data

        if config.WITH_CHAIN_ID_AND_TX_VERSION:
            ordered_fields["chainID"] = config.CHAIN_ID
            ordered_fields["version"] = config.TX_VERSION

        data_json = json.dumps(ordered_fields, separators=(',', ':')).encode("utf8")
        return data_json


class PreparedTransaction(PlainTransaction):
    def __init__(self, transaction=None, signature=None, dictionary=None):
        if dictionary:
            self._init_with_dictionary(dictionary)
        else:
            self._init_default(transaction, signature)

    def _init_with_dictionary(self, dictionary):
        self.__dict__.update(dictionary)

    def _init_default(self, transaction, signature):
        self.__dict__.update(transaction.payload())
        self.signature = signature

        if transaction.data:
            self.data = transaction.data

    @classmethod
    def from_file(cls, filename):
        data_json = utils.read_file(filename).encode()
        return cls.from_json(data_json)

    def save_to_file(self, filename):
        utils.write_file(filename, self.to_json())

    @classmethod
    def from_json(cls, json_data):
        dictionary = json.loads(json_data)
        return cls.from_dictionary(dictionary)

    def to_json(self):
        data_json = json.dumps(self.to_dictionary(), indent=4)
        return data_json

    @classmethod
    def from_dictionary(cls, dictionary):
        return cls(dictionary=dictionary)

    def to_dictionary(self):
        return self.__dict__.copy()

    def send(self, proxy):
        logger.info(f"PreparedTransaction.send:\n{self.to_json()}")
        tx_hash = proxy.send_transaction(self.to_dictionary())
        logger.info(f"Hash: {tx_hash}")
        return tx_hash


class BunchOfTransactions:
    def __init__(self):
        self.transactions = []

    def add_prepared(self, transaction):
        self.transactions.append(transaction)

    def add(self, sender, receiver_address, nonce, value, data, gas_price, gas_limit):
        plain = PlainTransaction()
        plain.nonce = int(nonce)
        plain.value = str(value)
        plain.sender = sender.address.bech32()
        plain.receiver = receiver_address
        plain.gasPrice = gas_price
        plain.gasLimit = gas_limit
        plain.data = data

        payload = TransactionPayloadToSign(plain)
        signature = signing.sign_transaction(payload, sender.pem_file)
        prepared = PreparedTransaction(plain, signature)

        self.transactions.append(prepared)

    def add_in_sequence(self):
        pass

    def send(self, proxy):
        logger.info(f"BunchOfTransactions.send: {len(self.transactions)} transactions")
        payload = [transaction.to_dictionary() for transaction in self.transactions]
        num_sent, hashes = proxy.send_transactions(payload)
        logger.info(f"Sent: {num_sent}")
        logger.info(f"TxsHashes: {hashes}")
        return num_sent, hashes


def prepare(args):
    workspace = args.workspace
    utils.ensure_folder(workspace)

    prepared = do_prepare_transaction(args)
    prepared_filename = path.join(workspace, f"tx-{args.tag}.json")
    prepared.save_to_file(prepared_filename)
    logger.info(f"Saved prepared transaction to {prepared_filename}")


def do_prepare_transaction(args):
    # "sender" taken from the PEM file
    sender_bytes = pem.get_pubkey(args.pem)
    sender_bech32 = Address(sender_bytes).bech32()

    plain = PlainTransaction()
    plain.nonce = int(args.nonce)
    plain.value = args.value
    plain.sender = sender_bech32
    plain.receiver = args.receiver
    plain.gasPrice = int(args.gas_price)
    plain.gasLimit = int(args.gas_limit)
    plain.data = args.data

    payload = TransactionPayloadToSign(plain)
    signature = signing.sign_transaction(payload, args.pem)
    prepared = PreparedTransaction(plain, signature)
    return prepared



import binascii
import glob
import logging
from os import path
from pathlib import Path

from erdpy import dependencies, errors, myprocess, utils
import shutil

logger = logging.getLogger("Project")


class Project:
    def __init__(self, directory):
        self.directory = str(Path(directory).resolve())

    def build(self, options=None):
        self.options = options or dict()
        self.debug = self.options.get("debug", False)
        self._ensure_dependencies_installed()
        self.perform_build()
        self._copy_build_artifacts_to_output()
        self._create_deploy_files()

    def _ensure_dependencies_installed(self):
        module_keys = self.get_dependencies()
        for module_key in module_keys:
            dependencies.install_module(module_key)

    def get_dependencies(self):
        raise NotImplementedError()

    def perform_build(self):
        raise NotImplementedError()

    def get_file_wasm(self):
        return self.find_file_in_output("*.wasm")

    def find_file_globally(self, pattern):
        folder = self.directory
        return self.find_file_in_folder(folder, pattern)

    def find_file_in_output(self, pattern):
        folder = path.join(self.directory, "output")
        return self.find_file_in_folder(folder, pattern)

    def find_file_in_folder(self, folder, pattern):
        files = list(Path(folder).rglob(pattern))

        if len(files) == 0:
            raise errors.KnownError(f"No file matches pattern [{pattern}].")
        if len(files) > 1:
            logging.warning(f"More files match pattern [{pattern}]. Will pick first:\n{files}")

        file = path.join(folder, files[0])
        return Path(file).resolve()

    def _copy_build_artifacts_to_output(self):
        raise NotImplementedError()

    def _copy_to_output(self, file):
        output_dir = path.join(self.directory, "output")
        utils.ensure_folder(output_dir)
        shutil.copy(file, output_dir)

    def _create_deploy_files(self):
        file_wasm = self.get_file_wasm()
        file_wasm_hex = file_wasm.with_suffix(".hex")

        with open(file_wasm, "rb") as file:
            bytecode_hex = binascii.hexlify(file.read())
        with open(file_wasm_hex, "wb+") as file:
            file.write(bytecode_hex)

    def get_bytecode(self):
        bytecode = utils.read_file(self.get_file_wasm().with_suffix(".hex"))
        return bytecode

    def run_tests(self, tests_directory, wildcard=None):
        testrunner_module = dependencies.get_module_by_key("arwentools")
        tool_directory = testrunner_module.get_directory()
        tool_env = testrunner_module.get_env()

        tool = path.join(tool_directory, "test")
        test_folder = path.join(self.directory, tests_directory)

        if not wildcard:
            args = [tool, test_folder]
            myprocess.run_process(args, env=tool_env)
        else:
            pattern = path.join(test_folder, wildcard)
            test_files = glob.glob(pattern)

            for test_file in test_files:
                print("Run test for:", test_file)
                args = [tool, test_file]
                myprocess.run_process(args, env=tool_env)



import logging
import os
from os import path

from erdpy import errors
from erdpy.wallet import bech32, pem, generate_pair
from erdpy import utils

logger = logging.getLogger("accounts")


class AccountsRepository:
    def __init__(self, folder):
        utils.ensure_folder(folder)
        self.folder = folder

    def get_account(self, name):
        pem_file = path.join(self.folder, f"{name}.pem")
        return Account(pem_file=pem_file)

    def generate_accounts(self, count):
        for i in range(count):
            self.generate_account(i)

    def generate_account(self, name):
        seed, pubkey = generate_pair()
        address = Address(pubkey).bech32()

        pem_file = f"{name}_{address}.pem"
        pem_file = path.join(self.folder, pem_file)
        pem.write(pem_file, seed, pubkey, name=f"{name}:{address}")

    def get_all(self):
        accounts = []
        for pem_file in os.listdir(self.folder):
            pem_file = path.join(self.folder, pem_file)
            account = Account(pem_file=pem_file)
            accounts.append(account)

        return accounts


class Account:
    def __init__(self, address=None, pem_file=None):
        self.private_key_seed = None
        self.address = Address(address)
        self.pem_file = pem_file
        self.nonce = 0

        if pem_file:
            seed, pubkey = pem.parse(pem_file)
            self.private_key_seed = seed.hex()
            self.address = Address(pubkey)

    def sync_nonce(self, proxy):
        logger.info(f"Account.sync_nonce()")
        self.nonce = proxy.get_account_nonce(self.address)
        logger.info(f"Account.sync_nonce() done: {self.nonce}")


class Address:
    HRP = "erd"
    PUBKEY_LENGTH = 32
    PUBKEY_STRING_LENGTH = PUBKEY_LENGTH * 2    # hex-encoded
    BECH32_LENGTH = 62

    def __init__(self, value):
        self._value_hex = None

        if not value:
            return

        # Copy-constructor
        if isinstance(value, Address):
            value = value._value_hex

        # We keep a hex-encoded string as the "backing" value
        if len(value) == Address.PUBKEY_LENGTH:
            self._value_hex = value.hex()
        elif len(value) == Address.PUBKEY_STRING_LENGTH:
            self._value_hex = _as_string(value)
        elif len(value) == Address.BECH32_LENGTH:
            self._value_hex = _decode_bech32(value).hex()
        else:
            raise errors.BadAddressFormatError(value)

    def hex(self):
        self._assert_validity()
        return self._value_hex

    def bech32(self):
        self._assert_validity()
        pubkey = self.pubkey()
        return bech32.bech32_encode(self.HRP, bech32.convertbits(pubkey, 8, 5))

    def pubkey(self):
        self._assert_validity()
        pubkey = bytes.fromhex(self._value_hex)
        return pubkey

    def _assert_validity(self):
        if self._value_hex is None:
            raise errors.EmptyAddressError()

    def __repr__(self):
        return self.bech32()

    @classmethod
    def zero(cls):
        return Address("0" * 64)


def _as_string(value):
    if isinstance(value, str):
        return value
    return value.decode("utf-8")


def _decode_bech32(value):
    bech32_string = _as_string(value)
    hrp, value_bytes = bech32.bech32_decode(bech32_string)
    if hrp != Address.HRP:
        raise errors.BadAddressFormatError(value)
    decodedBytes = bech32.convertbits(value_bytes, 5, 8, False)
    return bytearray(decodedBytes)



from erdpy.dependencies.install import install_group, install_module, get_install_directory, get_module_by_key, is_installed

__all__ = ["install_group", "install_module", "get_install_directory", "get_module_by_key", "is_installed"]



import logging

from erdpy.dependencies.config import get_all_modules

logger = logging.getLogger("install")


def install_group(group_name, overwrite=False):
    logger.info(f"install_group.group_name: {group_name}")

    modules = _get_modules_by_group(group_name)
    for module in modules:
        module.install(overwrite)


def install_module(key, overwrite=False):
    module = get_module_by_key(key)
    module.install(overwrite)


def get_install_directory(key):
    module = get_module_by_key(key)
    directory = module.get_directory()
    return directory


def _get_modules_by_group(group):
    return [module for module in get_all_modules() if group in module.groups]


def get_module_by_key(key):
    return [module for module in get_all_modules() if module.key == key][0]


def is_installed(group_name, overwrite=False) -> bool:
    modules = _get_modules_by_group(group_name)
    for module in modules:
        if not module.is_installed():
            return False
    return True



from Cryptodome.Hash import keccak

from erdpy import config, errors
from erdpy.accounts import Address
from erdpy.transactions import (PlainTransaction, PreparedTransaction,
                                TransactionPayloadToSign)
from erdpy.wallet import signing

VM_TYPE_ARWEN = "0500"


class SmartContract:
    def __init__(self, address=None, bytecode=None, metadata=None):
        self.address = Address(address)
        self.bytecode = bytecode
        self.metadata = metadata or CodeMetadata()

    def deploy(self, proxy, owner, arguments, gas_price, gas_limit, value):
        self.owner = owner
        self.owner.sync_nonce(proxy)
        self.compute_address()
        transaction = self.prepare_deploy_transaction(owner, arguments, gas_price, gas_limit, value)
        tx_hash = transaction.send(proxy)
        return tx_hash, self.address

    def prepare_deploy_transaction(self, owner, arguments, gas_price, gas_limit, value):
        arguments = arguments or []
        gas_price = int(gas_price or config.DEFAULT_GAS_PRICE)
        gas_limit = int(gas_limit)
        value = str(value or "0")

        plain = PlainTransaction()
        plain.nonce = owner.nonce
        plain.value = value
        plain.sender = owner.address.bech32()
        plain.receiver = Address.zero().bech32()
        plain.gasPrice = gas_price
        plain.gasLimit = gas_limit
        plain.data = self.prepare_deploy_transaction_data(arguments)

        payload = TransactionPayloadToSign(plain)
        signature = signing.sign_transaction(payload, owner.pem_file)
        prepared = PreparedTransaction(plain, signature)
        return prepared

    def prepare_deploy_transaction_data(self, arguments):
        tx_data = f"{self.bytecode}@{VM_TYPE_ARWEN}@{self.metadata.to_hex()}"

        for arg in arguments:
            tx_data += f"@{_prepare_argument(arg)}"

        return tx_data

    def compute_address(self):
        """
        8 bytes of zero + 2 bytes for VM type + 20 bytes of hash(owner) + 2 bytes of shard(owner)
        """
        owner_bytes = self.owner.address.pubkey()
        nonce_bytes = self.owner.nonce.to_bytes(8, byteorder="little")
        bytes_to_hash = owner_bytes + nonce_bytes
        address = keccak.new(digest_bits=256).update(bytes_to_hash).digest()
        address = bytes([0] * 8) + bytes([5, 0]) + address[10:30] + owner_bytes[30:]
        self.address = Address(address)

    def execute(self, proxy, caller, function, arguments, gas_price, gas_limit, value):
        self.caller = caller
        self.caller.sync_nonce(proxy)
        transaction = self.prepare_execute_transaction(caller, function, arguments, gas_price, gas_limit, value)
        tx_hash = transaction.send(proxy)
        return tx_hash

    def prepare_execute_transaction(self, caller, function, arguments, gas_price, gas_limit, value):
        arguments = arguments or []
        gas_price = int(gas_price or config.DEFAULT_GAS_PRICE)
        gas_limit = int(gas_limit)
        value = str(value or "0")

        plain = PlainTransaction()
        plain.nonce = caller.nonce
        plain.value = value
        plain.sender = caller.address.bech32()
        plain.receiver = self.address.bech32()
        plain.gasPrice = gas_price
        plain.gasLimit = gas_limit
        plain.data = self.prepare_execute_transaction_data(function, arguments)

        payload = TransactionPayloadToSign(plain)
        signature = signing.sign_transaction(payload, caller.pem_file)
        prepared = PreparedTransaction(plain, signature)
        return prepared

    def prepare_execute_transaction_data(self, function, arguments):
        tx_data = function

        for arg in arguments:
            tx_data += f"@{_prepare_argument(arg)}"

        return tx_data

    def upgrade(self, proxy, caller, arguments, gas_price, gas_limit, value):
        self.caller = caller
        self.caller.sync_nonce(proxy)
        transaction = self.prepare_upgrade_transaction(caller, arguments, gas_price, gas_limit, value)
        tx_hash = transaction.send(proxy)
        return tx_hash

    def prepare_upgrade_transaction(self, owner, arguments, gas_price, gas_limit, value):
        arguments = arguments or []
        gas_price = int(gas_price or config.DEFAULT_GAS_PRICE)
        gas_limit = int(gas_limit)
        value = str(value or "0")

        plain = PlainTransaction()
        plain.nonce = owner.nonce
        plain.value = value
        plain.sender = owner.address.bech32()
        plain.receiver = self.address.bech32()
        plain.gasPrice = gas_price
        plain.gasLimit = gas_limit
        plain.data = self.prepare_upgrade_transaction_data(arguments)

        payload = TransactionPayloadToSign(plain)
        signature = signing.sign_transaction(payload, owner.pem_file)
        prepared = PreparedTransaction(plain, signature)
        return prepared

    def prepare_upgrade_transaction_data(self, arguments):
        tx_data = f"upgradeContract@{self.bytecode}@{self.metadata.to_hex()}"

        for arg in arguments:
            tx_data += f"@{_prepare_argument(arg)}"

        return tx_data

    def query(self, proxy, function, arguments):
        arguments = arguments or []
        prepared_arguments = [_prepare_argument(argument) for argument in arguments]

        # TODO: move to proxy/core.py?
        payload = {
            "ScAddress": self.address.bech32(),
            "FuncName": function,
            "Args": prepared_arguments
        }

        response = proxy.query_contract(payload)
        return response["data"]["ReturnData"]


def _prepare_argument(argument):
    hex_prefix = "0X"
    as_string = str(argument).upper()

    if as_string.startswith(hex_prefix):
        return as_string[len(hex_prefix):]

    if not as_string.isnumeric():
        raise errors.UnknownArgumentFormat(as_string)

    as_number = int(as_string)
    as_hexstring = hex(as_number)[len(hex_prefix):]
    if len(as_hexstring) % 2 == 1:
        as_hexstring = "0" + as_hexstring

    return as_hexstring


class CodeMetadata:
    def __init__(self, upgradeable=False):
        self.upgradeable = upgradeable

    def to_hex(self):
        if self.upgradeable:
            return "0100"
        return "0000"



import logging
import signal
from os import path

from erdpy.contracts import _prepare_argument

from erdpy import config, dependencies, errors, myprocess, utils

logger = logging.getLogger("nodedebug")


def start(force=False):
    install_if_missing()

    if _is_running() and not force:
        return

    if force:
        stop()

    logger.info("Starting nodedebug...")
    myprocess.run_process_async(_get_args(), env=_get_env())


def _get_args():
    directory = dependencies.get_module_by_key("nodedebug").get_directory()
    executable = path.join(directory, "nodedebug")
    config_path = path.join(directory, "config", "config.toml")
    genesis_path = path.join(directory, "config", "genesis.json")

    args = [executable, "--rest-api-port", "8080", "--config",
            config_path, "--genesis-file", genesis_path]
    return args


async def start_async():
    await myprocess.async_subprocess(_get_args(), env=_get_env(), sinks=["nodedebug"])


def _is_running():
    proc = _get_proc()
    return proc is not None


def _get_proc():
    # TODO: Find another way to get the process, so that psutil is not needed
    # psutil requires "sudo apt-get install python3-dev" / "build-essentials"
    from psutil import process_iter

    for proc in process_iter():
        if proc.name() == "nodedebug":
            return proc


def install_if_missing():
    dependencies.install_module("nodedebug", overwrite=False)


def stop():
    proc = _get_proc()
    if proc:
        logger.info("Stopping nodedebug...")
        proc.send_signal(signal.SIGTERM)
        logger.info("Stopped nodedebug.")
    else:
        logger.info("Nodedebug wasn't started.")


def deploy(bytecode, owner, arguments=None, gas_price=None, gas_limit=None):
    logger.debug("deploy")

    arguments = arguments or []
    gas_limit = gas_limit
    gas_price = gas_price or config.DEFAULT_GAS_PRICE

    url = _get_url("deploy")

    tx_data = bytecode
    for arg in arguments:
        tx_data += f"@{_prepare_argument(arg)}"

    data = {
        "SndAddress": owner.address,
        "Value": "0",
        "GasLimit": gas_limit,
        "GasPrice": gas_price,
        "TxData": tx_data
    }

    raw_response = utils.post_json(url, data)
    logger.debug("Response: %s", raw_response)
    response = _Response(raw_response)
    response.verify()
    return response.tx_hash, response.contract_address


def execute(contract_address, caller, function, arguments=None, gas_price=None, gas_limit=None):
    logger.debug(f"execute, address={contract_address}")

    arguments = arguments or []
    gas_limit = gas_limit
    gas_price = gas_price or config.DEFAULT_GAS_PRICE

    url = _get_url("run")

    tx_data = function
    for arg in arguments:
        tx_data += f"@{_prepare_argument(arg)}"

    data = {
        "SndAddress": caller.address,
        "ScAddress": contract_address,
        "Value": "0",
        "GasLimit": gas_limit,
        "GasPrice": gas_price,
        "TxData": tx_data
    }

    raw_response = utils.post_json(url, data)
    logger.debug("Response: %s", raw_response)
    response = _VMOutputResponse(raw_response)
    response.verify()
    return response


def query(contract_address, function, arguments=None):
    logger.debug(f"query, address={contract_address}")

    arguments = [_prepare_argument(arg) for arg in arguments or []]

    url = _get_url("query")

    data = {
        "ScAddress": contract_address,
        "FuncName": function,
        "Args": arguments
    }

    raw_response = utils.post_json(url, data)
    logger.debug("Response: %s", raw_response)
    response = _VMOutputResponse(raw_response)
    response.verify()
    return response.return_data


def _get_url(action):
    return f"http://localhost:8080/vm-values/{action}"


class _Response:
    def __init__(self, raw_dict):
        def find(key):
            return utils.find_in_dictionary(raw_dict, key)

        self.error = find("error")
        self.contract_address = find("data.Address")
        self.tx_hash = find("data.Other.txHash")

    def verify(self):
        if self.error:
            raise errors.KnownError(self.error)


class _VMOutputResponse:
    def __init__(self, raw_dict):
        def find(key):
            return utils.find_in_dictionary(raw_dict, key)

        self.error = find("error")
        self.return_data = find("data.ReturnData")
        self.return_code = find("data.ReturnCode")

    def verify(self):
        if self.error:
            raise errors.KnownError(self.error)        


def _get_env():
    return dependencies.get_module_by_key("nodedebug").get_env()



import binascii
import logging

from erdpy.accounts import Address
from erdpy.config import MIN_GAS_LIMIT, GAS_PER_DATA_BYTE, MetaChainSystemSCsCost

logger = logging.getLogger("validators")

_STAKE_SMART_CONTRACT_ADDRESS = "erd1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqplllst77y4l"


def estimate_system_sc_call(args, base_cost, factor=1):
    num_bytes = len(args.data)
    gas_limit = MIN_GAS_LIMIT + num_bytes * GAS_PER_DATA_BYTE
    gas_limit += factor * base_cost
    return gas_limit


def parse_args_for_stake(args):
    num_of_nodes = int(args.number_of_nodes)
    keys = args.nodes_public_keys.split(',')
    if num_of_nodes != len(keys):
        logger.info("number of nodes is not equals with number of provided validators keys")
        return

    reward_address = args.reward_address

    stake_data = 'stake@' + binascii.hexlify(num_of_nodes.to_bytes(1, byteorder="little")).decode()
    for key in keys:
        stake_data += '@' + key + '@' + convert_to_hex('genesis')

    if reward_address:
        reward_address = Address(args.reward_address)
        stake_data += '@' + reward_address.hex()

    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS
    args.data = stake_data

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.STAKE, len(keys))

    return args


def parse_args_for_un_stake(args):
    parsed_keys, num_keys = parse_keys(args.nodes_public_keys)
    args.data = 'unStake' + parsed_keys
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNSTAKE, num_keys)

    return args


def parse_args_for_un_bond(args):
    parsed_keys, num_keys = parse_keys(args.nodes_public_keys)
    args.data = 'unBond' + parsed_keys
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNBOND, num_keys)

    return args


def parse_args_for_un_jail(args):
    parsed_keys, num_keys = parse_keys(args.nodes_public_keys)
    args.data = 'unJail' + parsed_keys
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNJAIL, num_keys)

    return args


def parse_args_for_changing_reward_address(args):
    reward_address = Address(args.reward_address)
    args.data = 'changeRewardAddress@' + reward_address.hex()
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.CHANGE_REWARD_ADDRESS)

    return args


def parse_args_for_claim(args):
    args.data = 'claim'
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.CLAIM)

    return args


def parse_keys(bls_public_keys):
    keys = bls_public_keys.split(',')
    parsed_keys = ''
    for key in keys:
        parsed_keys += '@' + key
    return parsed_keys, len(keys)


def convert_to_hex(key):
    return binascii.hexlify(bytes(key, 'utf-8')).decode()



import sys

from erdpy import config, facade


def setup_parser(subparsers):
    parser = subparsers.add_parser("validator")
    subparsers = parser.add_subparsers()

    sub = subparsers.add_parser("stake")
    _add_common_arguments(sub)
    sub.add_argument("--number-of-nodes", required=True)
    sub.add_argument("--nodes-public-keys", required=True)
    sub.add_argument("--reward-address", default="")
    sub.set_defaults(func=do_stake)

    sub = subparsers.add_parser("unstake")
    _add_common_arguments(sub)
    sub.add_argument("--nodes-public-keys", required=True)
    sub.set_defaults(func=do_unstake)

    sub = subparsers.add_parser("unjail")
    _add_common_arguments(sub)
    sub.add_argument("--nodes-public-keys", required=True)
    sub.set_defaults(func=do_unjail)

    sub = subparsers.add_parser("unbond")
    _add_common_arguments(sub)
    sub.add_argument("--nodes-public-keys", required=True)
    sub.set_defaults(func=do_unbond)

    sub = subparsers.add_parser("change-reward-address")
    _add_common_arguments(sub)
    sub.add_argument("--reward-address", required=True)
    sub.set_defaults(func=change_reward_address)

    sub = subparsers.add_parser("claim")
    _add_common_arguments(sub)
    sub.set_defaults(func=do_claim)


def _add_common_arguments(sub):
    sub.add_argument("--pem", required=True)
    sub.add_argument("--nonce", type=int, required=not("--recall-nonce" in sys.argv))
    sub.add_argument("--recall-nonce", action="store_true", default=False)
    sub.add_argument("--value", default="0")
    sub.add_argument("--gas-price", default=config.DEFAULT_GAS_PRICE)
    sub.add_argument("--gas-limit", required=not("--estimate-gas" in sys.argv))
    sub.add_argument("--estimate-gas", action="store_true", default=False)
    sub.add_argument("--proxy", required=True)


def do_stake(args):
    facade.prepare_and_send_stake_transaction(args)


def do_unstake(args):
    facade.prepare_and_send_unstake_transaction(args)


def do_unbond(args):
    facade.prepare_and_send_unbond_transaction(args)


def do_unjail(args):
    facade.prepare_and_send_unjail_transaction(args)


def change_reward_address(args):
    facade.prepare_and_send_change_reward_address_transaction(args)


def do_claim(args):
    facade.prepare_and_send_claim_transaction(args)



from erdpy.projects.templates_repository import TemplatesRepository


def get_templates_repositories():
    return [
        TemplatesRepository(
            key="sc-examples",
            url="https://github.com/ElrondNetwork/sc-examples/archive/master.zip",
            github="ElrondNetwork/sc-examples",
            relative_path="sc-examples-master"
        ),

        TemplatesRepository(
            key="sc-examples-rs",
            url="https://github.com/ElrondNetwork/sc-examples-rs/archive/master.zip",
            github="ElrondNetwork/sc-examples-rs",
            relative_path="sc-examples-rs-master"
        )
    ]



from erdpy.dependencies.modules import ArwenToolsModule, Rust, StandaloneModule


def get_all_modules():
    return [
        StandaloneModule(key="llvm", name="llvm", tag="v9", groups=["clang", "cpp"],
                         urls_by_platform={
                             "linux": "vendor-llvm/v9-19feb/linux-amd64.tar.gz?t=19feb",
                             "osx": "vendor-llvm/v9-19feb/darwin-amd64.tar.gz",
                             "windows": "vendor-llvm/v9-19feb/windows-amd64.tar.gz",
        }),

        ArwenToolsModule(key="arwentools", name="arwentools", tag="g03d0b3f", groups=["arwentools"],
                         urls_by_platform={
                             "linux": "travis-builds/ARWEN_v0.3.20-24-ge8c5c68_linux_amd64.tar.gz",
                             "osx": "travis-builds/ARWEN_v0.3.20-24-ge8c5c68_darwin_amd64.tar.gz"
        }),

        Rust(key="rust", name="rust", tag="initial", groups=["rust"])
    ]



import logging

from erdpy import utils, wallet
from erdpy.accounts import Account, Address
from erdpy.blockatlas import BlockAtlas
from erdpy.contracts import CodeMetadata, SmartContract
from erdpy.dispatcher.transactions.queue import TransactionQueue
from erdpy.environments import TestnetEnvironment
from erdpy.projects import load_project
from erdpy.proxy import ElrondProxy, TransactionCostEstimator
from erdpy.transactions import PreparedTransaction, do_prepare_transaction
from erdpy.validators import validators
from erdpy.wallet import pem

logger = logging.getLogger("facade")


def deploy_smart_contract(args):
    logger.debug("deploy_smart_contract")

    project_directory = args.project
    pem_file = args.pem
    proxy_url = args.proxy
    arguments = args.arguments
    gas_price = args.gas_price
    gas_limit = args.gas_limit
    value = args.value
    metadata_upgradeable = args.metadata_upgradeable

    # TODO: apply guards

    project = load_project(project_directory)
    bytecode = project.get_bytecode()
    metadata = CodeMetadata(metadata_upgradeable)
    contract = SmartContract(bytecode=bytecode, metadata=metadata)
    environment = TestnetEnvironment(proxy_url)
    owner = Account(pem_file=pem_file)

    def flow():
        tx_hash, address = environment.deploy_contract(contract, owner, arguments, gas_price, gas_limit, value)
        logger.info("Tx hash: %s", tx_hash)
        logger.info("Contract address: %s", address)
        utils.dump_out_json({"tx": tx_hash, "contract": address.bech32()}, args.outfile)

    environment.run_flow(flow)


def call_smart_contract(args):
    logger.debug("call_smart_contract")

    contract_address = args.contract
    pem_file = args.pem
    proxy_url = args.proxy
    function = args.function
    arguments = args.arguments
    gas_price = args.gas_price
    gas_limit = args.gas_limit
    value = args.value

    contract = SmartContract(contract_address)
    environment = TestnetEnvironment(proxy_url)
    caller = Account(pem_file=pem_file)

    def flow():
        tx_hash = environment.execute_contract(contract, caller, function, arguments, gas_price, gas_limit, value)
        logger.info("Tx hash: %s", tx_hash)

    environment.run_flow(flow)


def upgrade_smart_contract(args):
    logger.debug("upgrade_smart_contract")

    contract_address = args.contract
    project_directory = args.project
    pem_file = args.pem
    proxy_url = args.proxy
    arguments = args.arguments
    gas_price = args.gas_price
    gas_limit = args.gas_limit
    value = args.value
    metadata_upgradeable = args.metadata_upgradeable

    project = load_project(project_directory)
    bytecode = project.get_bytecode()
    metadata = CodeMetadata(metadata_upgradeable)
    contract = SmartContract(contract_address, bytecode=bytecode, metadata=metadata)
    environment = TestnetEnvironment(proxy_url)
    caller = Account(pem_file=pem_file)

    def flow():
        tx_hash = environment.upgrade_contract(contract, caller, arguments, gas_price, gas_limit, value)
        logger.info("Tx hash: %s", tx_hash)

    environment.run_flow(flow)


def query_smart_contract(args):
    logger.debug("query_smart_contract")

    contract_address = args.contract
    proxy_url = args.proxy
    function = args.function
    arguments = args.arguments

    contract = SmartContract(contract_address)
    environment = TestnetEnvironment(proxy_url)

    def flow():
        result = environment.query_contract(contract, function, arguments)
        print(result)

    environment.run_flow(flow)


def get_account_nonce(args):
    proxy_url = args.proxy
    address = args.address

    proxy = ElrondProxy(proxy_url)
    nonce = proxy.get_account_nonce(Address(address))
    print(nonce)
    return nonce


def get_account_balance(args):
    proxy_url = args.proxy
    address = args.address

    proxy = ElrondProxy(proxy_url)
    balance = proxy.get_account_balance(Address(address))
    print(balance)
    return balance


def get_account(args):
    proxy_url = args.proxy
    address = args.address

    proxy = ElrondProxy(proxy_url)
    account = proxy.get_account(Address(address))
    print(account)
    return account


def get_account_transactions(args):
    proxy_url = args.proxy
    address = args.address

    proxy = ElrondProxy(proxy_url)
    account = proxy.get_account_transactions(Address(address))
    print(account)
    return account


def get_num_shards(args):
    proxy_url = args.proxy
    proxy = ElrondProxy(proxy_url)
    num_shards = proxy.get_num_shards()
    print(num_shards)
    return num_shards


def get_last_block_nonce(args):
    proxy_url = args.proxy
    shard_id = args.shard_id
    proxy = ElrondProxy(proxy_url)
    nonce = proxy.get_last_block_nonce(shard_id)
    print(nonce)
    return nonce


def get_gas_price(args):
    proxy_url = args.proxy
    proxy = ElrondProxy(proxy_url)
    price = proxy.get_gas_price()
    print(price)
    return price


def get_chain_id(args):
    proxy_url = args.proxy
    proxy = ElrondProxy(proxy_url)
    chain_id = proxy.get_chain_id()
    print(chain_id)
    return chain_id


def get_meta_nonce(args):
    proxy = ElrondProxy(args.proxy)
    nonce = proxy.get_meta_nonce()
    print(nonce)
    return nonce


def get_meta_block(args):
    proxy = ElrondProxy(args.proxy)
    block = proxy.get_meta_block(args.nonce)
    print(block)
    return block


def get_transaction_cost(args):
    logger.debug("call_get_transaction_cost")

    cost_estimator = TransactionCostEstimator(args.proxy)
    result = cost_estimator.estimate_tx_cost(args)
    print(result)
    return result


def send_prepared_transaction(args):
    proxy = ElrondProxy(args.proxy)
    prepared = PreparedTransaction.from_file(args.tx)
    tx_hash = prepared.send(proxy)
    print(tx_hash)
    return tx_hash


def prepare_and_send_transaction(args):
    proxy = ElrondProxy(args.proxy)

    if args.recall_nonce:
        owner = Account(pem_file=args.pem)
        owner.sync_nonce(proxy)
        args.nonce = owner.nonce

    prepared = do_prepare_transaction(args)
    tx_hash = prepared.send(proxy)
    print(tx_hash)
    return tx_hash


def prepare_and_send_stake_transaction(args):
    args = validators.parse_args_for_stake(args)
    return prepare_and_send_transaction(args)


def prepare_and_send_unstake_transaction(args):
    args = validators.parse_args_for_un_stake(args)
    return prepare_and_send_transaction(args)


def prepare_and_send_unjail_transaction(args):
    args = validators.parse_args_for_un_jail(args)
    return prepare_and_send_transaction(args)


def prepare_and_send_unbond_transaction(args):
    args = validators.parse_args_for_un_bond(args)
    return prepare_and_send_transaction(args)


def prepare_and_send_change_reward_address_transaction(args):
    args = validators.parse_args_for_changing_reward_address(args)
    return prepare_and_send_transaction(args)


def prepare_and_send_claim_transaction(args):
    args = validators.parse_args_for_claim(args)
    return prepare_and_send_transaction(args)


def enqueue_transaction(args):
    queue = TransactionQueue()
    queue.enqueue_transaction(args)


def dispatch_transactions(args):
    queue = TransactionQueue()
    queue.dispatch_transactions(args)


def dispatch_transactions_continuously(args):
    queue = TransactionQueue()
    queue.dispatch_transactions_continuously(args)


def clean_transactions_queue():
    queue = TransactionQueue()
    queue.clean_transactions_queue()


def generate_pem(args):
    pem_file = args.pem
    mnemonic = args.mnemonic

    seed, pubkey = wallet.generate_pair()
    if mnemonic:
        mnemonic = input("Enter mnemonic:\n")
        seed, pubkey = wallet.derive_keys(mnemonic)

    address = Address(pubkey)
    pem.write(pem_file, seed, pubkey, name=address.bech32())
    logger.info(f"Created PEM file [{pem_file}] for [{address.bech32()}]")


def do_bech32(args):
    encode = args.encode
    value = args.value
    address = Address(value)

    result = address.bech32() if encode else address.hex()
    print(result)
    return result


def blockatlas_get_current_block_number(args):
    client = BlockAtlas(args.url, args.coin)
    number = client.get_current_block_number()
    print(number)
    return number


def blockatlas_get_block_by_number(args):
    client = BlockAtlas(args.url, args.coin)
    block = client.get_block_by_number(args.number)
    print(block)
    return block


def blockatlas_get_txs_by_address(args):
    client = BlockAtlas(args.url, args.coin)
    transactions = client.get_txs_by_address(args.address)
    print(transactions)
    return transactions



import logging
import nacl.encoding
import nacl.signing

from erdpy.wallet import pem

logger = logging.getLogger("wallet")


def sign_transaction(transaction, pem_file):
    seed, _ = pem.parse(pem_file)
    signing_key = nacl.signing.SigningKey(seed)

    data_json = transaction.to_json()
    signed = signing_key.sign(data_json)
    signature = signed.signature
    signature_hex = signature.hex()

    return signature_hex



import logging

from erdpy import dependencies, errors

logger = logging.getLogger("cli.deps")


def setup_parser(subparsers):
    parser = subparsers.add_parser("deps")
    subparsers = parser.add_subparsers()

    choices = ["clang", "cpp", "rust", "arwentools"]

    sub = subparsers.add_parser("install", description="Install dependencies or elrond-sdk packages.")
    sub.add_argument("group", choices=choices, help="the dependency to install")
    sub.add_argument("--overwrite", action="store_true", default=False)
    sub.set_defaults(func=install)

    sub = subparsers.add_parser("check", description="Check whether a dependency is installed.")
    sub.add_argument("group", choices=choices, help="the dependency to check")
    sub.set_defaults(func=check)


def install(args):
    group = args.group
    overwrite = args.overwrite
    dependencies.install_group(group, overwrite=overwrite)


def check(args):
    group = args.group
    installed = dependencies.is_installed(group)
    if installed:
        logger.info(f"[{group}] is installed.")
        return
    raise errors.DependencyMissing(group)
